import pickle
import pprint
from weakref import finalize
from pathlib import Path
from typing import Generator, List, Union, Iterator, Callable

DataSet = Union[Iterator, Callable[..., Generator[dict, None, None]]]


class Data:
    def __init__(self, data: DataSet, workflow: List[Callable] = None, cache=False):
        self._data = data
        self._workflow = [] if workflow is None else workflow
        self._cache_status: bool = cache
        self._len = None
        self._finalizer = finalize(self, self.remove)

    def remove(self):
        filename = f"{str(id(self))}.pickle"
        if self.__check_cache(filename):
            path = Path("./").joinpath("temp").joinpath(filename)
            path.unlink(missing_ok=True)
        del self._data

    def __iter__(self) -> DataSet:
        filename = f"{str(id(self))}.pickle"

        if self._cache_status and self.__check_cache(filename):
            working_data = self.__load_file(filename)
            for record in working_data:
                yield record
        else:
            file = None
            if self._cache_status:
                filepath = f"./temp/{filename}"
                file = open(filepath, "wb")
            for record in self.__apply_workflow():
                if file is not None:
                    pickle.dump(record, file)
                yield record

            if file:
                file.close()

    def __len__(self):
        if self._len is not None:
            return self._len
        else:
            self._len = 0
            for _ in self:
                self._len += 1
            return self._len

    def __check_cache(self, filename: str) -> bool:
        """Checks whether file exist.

        :param filename: Filename.
        :return: File exists or not.
        """
        path = Path("./").joinpath("temp")
        path.mkdir(exist_ok=True)
        path = path.joinpath(filename)
        return path.is_file()

    def __load_file(self, filename: str) -> Generator[dict, None, None]:
        """Loads records from pickle file.

        :param filename: Filepath.
        :return: Generator records.
        """
        path = Path("./").joinpath("temp").joinpath(filename)
        if not path.is_file():
            raise ValueError(f"{filename} isn't file.")

        if path.suffix != ".pickle":
            raise ValueError(f"File hasn't pickle extension.")

        with open(path.resolve(), "rb") as file:
            while True:
                try:
                    decoded_data = pickle.load(file)
                    yield decoded_data
                except EOFError:
                    break

    def __apply_workflow(self) -> Generator[dict, None, None]:
        """Creates generator records with apply workflow.

        :return: Generator records.
        """
        working_data = self._data() if callable(self._data) else self._data
        for record in working_data:
            for step in self._workflow:
                if isinstance(record, (list, tuple)):
                    record = [r for r in record if step["callback"](r) is not None]
                    if not record:
                        record = None
                        break
                else:
                    record = step["callback"](record)
                    if record is None:
                        record = None
                        break

            if record is not None:
                if isinstance(record, (list, tuple)):
                    for r in record:
                        yield r
                else:
                    yield record

    def filter(self, callback: Callable) -> "Data":
        """Append filter to workflow.

        :param callback: Filter function.
        """
        new_workflow = [*self._workflow.copy(), {"filter": True, "callback": lambda record: record if callback(record) else None}]
        return Data(self._data, new_workflow, self._cache_status)

    def map(self, callback: Callable) -> "Data":
        """Append transform function to workflow.

        :param callback: Transform function.
        """
        new_workflow = [*self._workflow.copy(), {"filter": False, "callback": callback}]
        return Data(self._data, new_workflow, self._cache_status)

    def sift(self, limit: int = None, skip: int = None) -> Generator[dict, None, None]:
        """Skips and limits records.

        :param limit: Limited records.
        :param skip: Skipped records.
        :return: Generator records.
        """
        skipped = 0
        pushed = 0

        for record in self:
            if skip is not None and skipped < skip:
                skipped += 1
                continue
            if limit is not None and pushed == limit:
                break
            yield record
            pushed += 1

    def use_cache(self, status: bool) -> "Data":
        """Change status cache.

        :param status: Status.
        """
        if status:
            self._cache_status = True
        else:
            self._cache_status = False

        return self

    def find_by(self, record_field, field_values) -> Generator:
        """Get the records whose field value is written in the field_values list."""
        values_for_find = list(field_values)
        for record in self:
            if values_for_find:
                if record[record_field] in values_for_find:
                    values_for_find.remove(record[record_field])
                    yield record
                else:
                    continue
            else:
                break

    def __str__(self):
        output = "------------- Printed first 5 records -------------\n"
        for record in self.sift(limit=5):
            output += pprint.pformat(record) + "\n"
        return output

    def __bool__(self):
        for _ in self:
            return True
        return False



import pickle
from functools import partial

import requests
import json
import urllib3
from weakref import finalize
from pathlib import Path
from datetime import datetime
from csv import DictReader
from pprint import pformat
from typing import Generator, Iterable, List, Union
from urllib.parse import urlencode, urlparse
from sseclient import SSEClient
from th2_data_services.data import Data


class DataSource:
    def __init__(self, url, chunk_length: int = 128):
        self.url = url
        self._finalizer = finalize(self, self.remove)
        self.__chunk_length = chunk_length

    def remove(self):
        filename = urlparse(self.__url).netloc
        path = Path("./").joinpath("temp")
        if path.exists():
            for file in path.iterdir():
                current_file = str(file)
                if filename in current_file:
                    file.unlink()

    @property
    def url(self):
        return self.__url

    @url.setter
    def url(self, url):
        if url[-1] == "/":
            url = url[:-1]
        self.__url = url

    def sse_request_to_data_provider(self, **kwargs) -> Generator[dict, None, None]:
        """Sends SSE request. For create custom sse-request to data-provider
        use this readme https://github.com/th2-net/th2-rpt-data-provider .

        :param kwargs: Query options.
        :return: SSE response data.
        """
        route = kwargs.get("route")
        if not route:
            raise ValueError("Route is required field. Please fill it.")

        if kwargs.get("startTimestamp") and isinstance(kwargs.get("startTimestamp"), datetime):
            kwargs["startTimestamp"] = int(kwargs["startTimestamp"].timestamp() * 1000)  # unix timestamp in milliseconds
        if kwargs.get("endTimestamp") and isinstance(kwargs.get("endTimestamp"), datetime):
            kwargs["endTimestamp"] = int(kwargs["endTimestamp"].timestamp() * 1000)

        url = self.__url + route
        url = f"{url}?{urlencode(kwargs)}"

        yield from self.__execute_sse_request(url)

    def get_events_from_data_provider(self, cache=False, **kwargs) -> Data:
        """Sends SSE request for events. For help use this readme
        https://github.com/th2-net/th2-rpt-data-provider#sse-requests-api
        on route http://localhost:8080/search/sse/events.

        :param cache: Flag if you what save to cache.
        :param kwargs: Query options.
        :return: Events.
        """
        if not kwargs.get("startTimestamp") and not kwargs.get("resumeFromId"):
            raise ValueError(
                "'startTimestamp' or 'resumeFromId' must not be null for route /search/sse/events. Please note it. "
                "More information on request here: https://github.com/th2-net/th2-rpt-data-provider"
            )
        if isinstance(kwargs["startTimestamp"], datetime):
            kwargs["startTimestamp"] = int(kwargs["startTimestamp"].timestamp() * 1000)  # unix timestamp in milliseconds

        if kwargs.get("endTimestamp") and isinstance(kwargs.get("endTimestamp"), datetime):
            kwargs["endTimestamp"] = int(kwargs["endTimestamp"].timestamp() * 1000)

        url = self.__url + "/search/sse/events"
        url = f"{url}?{urlencode(kwargs)}"

        filename = None
        if cache:
            filename = urlparse(self.__url).netloc + f"_events_{urlencode(kwargs)}"
            filename = f"{filename}.pickle"

        if filename and self.__check_cache(filename):
            data = self.__load_file(filename)
        else:
            data = partial(self.__load_from_provider, url, filename if cache else None)
        return Data(data)

    def get_messages_from_data_provider(self, cache: bool = False, **kwargs) -> Data:
        """Sends SSE request for messages. For help use this readme
        https://github.com/th2-net/th2-rpt-data-provider#sse-requests-api
        on route http://localhost:8080/search/sse/messages.

        :param cache: Flag if you what save to cache.
        :param kwargs: Query options.
        :return: Messages.
        """
        if not kwargs.get("startTimestamp") and not kwargs.get("resumeFromId"):
            raise ValueError(
                "'startTimestamp' or 'resumeFromId' must not be null for route /search/sse/messages. Please note it. "
                "More information on request here: https://github.com/th2-net/th2-rpt-data-provider"
            )
        if not kwargs.get("stream"):
            raise ValueError("'stream' is required field. Please note it." "More information on request here: https://github.com/th2-net/th2-rpt-data-provider")

        if isinstance(kwargs["startTimestamp"], datetime):
            kwargs["startTimestamp"] = int(kwargs["startTimestamp"].timestamp() * 1000)  # unix timestamp in milliseconds

        if kwargs.get("endTimestamp") and isinstance(kwargs.get("endTimestamp"), datetime):
            kwargs["endTimestamp"] = int(kwargs["endTimestamp"].timestamp() * 1000)

        streams = kwargs.pop("stream")
        if isinstance(streams, (list, tuple)):
            streams = f"&stream=".join(streams)
        streams = f"&stream={streams}"

        url = self.__url + "/search/sse/messages"
        url = f"{url}?{urlencode(kwargs) + streams}"

        filename = None
        if cache:
            filename = urlparse(self.__url).netloc + f"_messages_{urlencode(kwargs)}"
            filename = f"{filename}.pickle"

        if filename and self.__check_cache(filename):
            data = self.__load_file(filename)
        else:
            data = partial(self.__load_from_provider, url, filename)
        return Data(data)

    def __check_cache(self, filename: str) -> bool:
        """Checks whether file exist.

        :param filename: Filename.
        :return: File exists or not.
        """
        path = Path("./temp")
        path.mkdir(exist_ok=True)
        path = path.joinpath(filename)
        return path.is_file()

    def __load_file(self, filename: str) -> Generator[dict, None, None]:
        """Loads records from pickle file.

        :param filename: Filename.
        :return: Generator records.
        """
        path = Path("./").joinpath("temp").joinpath(filename)
        if not path.is_file():
            raise ValueError(f"{filename} isn't file.")

        if path.suffix != ".pickle":
            raise ValueError(f"File hasn't pickle extension.")

        with open(path, "rb") as file:
            while True:
                try:
                    decoded_data = pickle.load(file)
                    yield decoded_data
                except EOFError:
                    break

    def __load_from_provider(self, url: str, filename: str = None) -> Generator[dict, None, None]:
        """Loads records from data provider.

        :param url: Url.
        :param filename: Filename if you what to create local storage.
        :return: Generator records.
        """
        file = None
        if filename is not None:
            path = Path("./").joinpath("temp").joinpath(filename)
            file = open(path, "wb")

        for record in self.__execute_sse_request(url):
            if file is not None:
                pickle.dump(record, file)
            yield record

        if file:
            file.close()

    def __execute_sse_request(self, url: str):
        """Creates SSE connection to server.

        :param url: Url.
        :return: Response data.
        """
        response = self.__create_stream_connection(url)
        client = SSEClient(response)
        for record in client.events():
            if record.event not in ["close", "error", "keep_alive"]:
                record_data = json.loads(record.data)
                yield record_data

    def __create_stream_connection(self, url: str):
        """Create stream connection.

        :param url: Url.
        :return: Response stream data.
        """
        headers = {"Accept": "text/event-stream"}
        http = urllib3.PoolManager()
        response = http.request(method="GET", url=url, headers=headers, preload_content=False)

        for chunk in response.stream(self.__chunk_length):
            yield chunk

        response.release_conn()

    def find_messages_by_id_from_data_provider(self, messages_id: Union[Iterable, str]) -> Union[List[dict], dict]:
        """Gets messages by ids using URL request.

        :param messages_id: Messages id.
        :return: List[Message_dict] if you request a list or Message_dict.
        """
        if isinstance(messages_id, str):
            messages_id = [messages_id]
        result = []
        for msg_id in messages_id:
            response = requests.get(f"{self.__url}/message/{msg_id}")
            result.append(response.json())
        return result if len(result) > 1 else result[0]

    def find_events_by_id_from_data_provider(self, events_id: Union[Iterable, str]) -> Union[List[dict], dict]:
        """Gets events by ids using URL request.

        :param events_id: Events id.
        :return: List[Event_dict] if you request a list or Event_dict.
        """
        if isinstance(events_id, str):
            events_id = [events_id]
        events = "&".join([f"ids={id_}" for id_ in events_id])
        response = requests.get(f"{self.__url}/events/?{events}")
        answer = response.json()
        return answer if len(answer) > 1 else answer[0]

    @staticmethod
    def read_csv_file(*sources: str) -> Generator[str, None, None]:
        """Gets data from csv files.

        :param sources: Path to files.
        :return: Csv files payload.
        """
        for source in sources:
            with open(source) as data:
                for message in DictReader(data):
                    yield dict(message)

    @staticmethod
    def write_to_txt(data: Generator[str, None, None], source: str) -> None:
        """Writes to txt files.

        :param data: Data.
        :param source: Path to file.
        """
        with open(source, "w") as txt_file:
            for record in data:
                txt_file.write(f"{pformat(record)}\n" + ("-" * 50) + "\n")



from collections import defaultdict
from typing import Generator, Union, Iterator, Optional, Callable, Dict, List

from th2_data_services.data import Data
from th2_data_services.data_source import DataSource
from anytree import Node, RenderTree, findall


class EventsTree:
    """
    EventsTree - is a useful wrapper for your retrieved data.

    EventsTree - is not a tree in the literal sense.
    It is an object with a dict 'events' inside which contains
    events without their body.

    EventTree contains all events inside, so it takes
    ~2.5Gb for 1 million events.

    Note:
        Take a look at the following HTML tree to understand some important terms.

        <body> <!-- ancestor (grandparent), but not parent -->
            <div> <!-- parent & ancestor -->
                <p>Hello, world!</p> <!-- child -->
                <p>Goodbye!</p> <!-- sibling -->
            </div>
        </body>

    """

    def __init__(self, data: Union[Iterator, Generator[dict, None, None], Data] = None):
        if data is None:
            data = []

        self._events = {}  # {EventID_str: Event_dict}
        self._unknown_events = defaultdict(lambda: 0)  # {parent_id: int(cnt)}
        self.build_tree(data)

    @property
    def events(self) -> dict:
        return self._events

    @property
    def unknown_events(self):
        return self._unknown_events

    def clear_events(self) -> None:
        """Clear exist events."""
        self._events = {}

    def clear_unknown_events(self) -> None:
        """Clear unknown events."""
        self._unknown_events.clear()

    def build_tree(self, data: Union[Iterator, Generator[dict, None, None]]) -> None:
        """Build or append new events to family tree.

        :param data: Events.
        """
        for event in data:
            event = event.copy()
            self.append_element(event)
        self.search_unknown_parents()

    def append_element(self, event: dict) -> None:
        """Append new event to events tree.

        Will update the event if event_id matches.
        Will remove the event from unknown_events if it in unknown_events dict.

        :param event: Event
        """
        event_id = event["eventId"]
        try:
            event.pop("body")
        except KeyError:
            pass

        if ":" in event_id:
            # event_id sometimes looks like batchId:eventId
            event_id = event_id.split(":")[-1]
        self._events[event_id] = event

        event_id = event["eventId"]
        if event_id in self._unknown_events:
            self._unknown_events.pop(event_id)

    def search_unknown_parents(self) -> dict:
        """Searches unknown events.

        :return: Unknown events.
        """
        self.clear_unknown_events()
        event: dict
        for event in self.events.values():
            parent_id = event["parentEventId"]
            if parent_id is not None:
                if ":" in parent_id:
                    # parent_id sometimes looks like batchId:eventId
                    parent_id = parent_id.split(":")[-1]

                if parent_id not in self._events:
                    parent_id = event["parentEventId"]
                    self._unknown_events[parent_id] += 1

        return self._unknown_events

    def is_in_ancestor_name(self, event: dict, event_name: str):
        """Verify event has ancestor with specified event name.

        :param event: Event parent id.
        :param event_name: Event name.
        :return: True/False.
        """
        parent_id = event.get("parentEventId")

        if parent_id is None:
            raise ValueError("event must have field 'parentEventId'")

        ancestor = self._events.get(parent_id)
        while ancestor:
            if event_name in ancestor.get("eventName"):
                return True
            ancestor = self._events.get(ancestor.get("parentEventId"))
        return False

    def is_in_ancestor_type(self, event: dict, event_type: str) -> bool:
        """Verify event has ancestor with specified event type.

        :param event: Event.
        :param event_type: Event type.
        :return: True/False.
        """
        parent_id = event.get("parentEventId")

        if parent_id is None:
            raise ValueError("event must have field 'parentEventId'")

        ancestor = self._events.get(parent_id)
        while ancestor:
            if event_type == ancestor.get("eventType"):
                return True
            ancestor = self._events.get(ancestor.get("parentEventId"))
        return False

    def get_ancestor_by_name(self, event: dict, event_name: str) -> Optional[dict]:
        """Gets event ancestor by event_name.

        :param event: Record.
        :param event_name: Event name.
        :return: Event.
        """
        parent_id = event.get("parentEventId")

        if parent_id is None:
            raise ValueError("event must have field 'parentEventId'")

        ancestor = self._events.get(parent_id)
        while ancestor:
            if event_name in ancestor.get("eventName"):
                return ancestor
            ancestor = self._events.get(ancestor.get("parentEventId"))
        return ancestor

    def get_ancestor_by_super_type(
        self,
        event: dict,
        super_type: str,
        super_type_get_func: Callable[[dict, Dict[int, dict]], str],
    ) -> Optional[dict]:
        """Gets event ancestor by super_type.

        :param event: Event.
        :param super_type: Super type.
        :param super_type_get_func: Super type get function.
        :return: Event.
        """
        parent_id = event.get("parentEventId")

        if parent_id is None:
            raise ValueError("event must have field 'parentEventId'")

        if parent_id and ":" in parent_id:
            parent_id = parent_id.split(":")[-1]

        ancestor = self._events.get(parent_id)
        while ancestor:
            if super_type == super_type_get_func(ancestor, self._events):
                return ancestor
            parent_id = ancestor.get("parentEventId")
            if parent_id and ":" in parent_id:
                parent_id = parent_id.split(":")[-1]
            ancestor = self._events.get(parent_id)
        return ancestor

    def recover_unknown_events(self, data_source: DataSource) -> None:
        """Loads unknown events from data provider and recover EventsTree.

        :param data_source: DataSources.
        """
        old_unknown_events = self._unknown_events.keys()
        while self._unknown_events:
            new_events = data_source.find_events_by_id_from_data_provider(self._unknown_events.keys())
            if isinstance(new_events, dict):
                new_events = [new_events]
            self.build_tree(new_events)
            if self._unknown_events == old_unknown_events:
                break
            old_unknown_events = self._unknown_events.copy()

    def get_children(self, parent_event_id) -> list:
        return [e for e in self._events.values() if e["parentEventId"] == parent_event_id]


class TreeNode(Node):
    separator = " | "

    def __str__(self):
        s = ""
        for pre, fill, node in RenderTree(self):
            status = "P" if node.data["successful"] else "F"
            s += "[%s] %s%s\n" % (status, pre, node.name)
        return s

    def show(self, fmt: Callable = None, failed_only=False, show_status=True):
        if fmt is None:

            def fmt(pre, fill, node):
                if show_status:
                    status = "P" if node.data["successful"] else "F"
                    return "[%s] %s%s\n" % (status, pre, node.name)
                else:
                    return "%s%s\n" % (pre, node.name)

        s = ""
        if not failed_only:
            for pre, fill, node in RenderTree(self):
                s += fmt(pre, fill, node)
        else:
            for pre, fill, node in RenderTree(self):
                if not node.data["successful"]:
                    s += fmt(pre, fill, node)
        return s

    def get_by_status(self, status: bool):
        return findall(self, lambda node: node.data["successful"] is status)

    def get_by_leaves_status(self, status: bool):
        return [n for n in self.leaves if n.data["successful"] is status]


class EventsTree2:
    """
    EventsTree2 - experimental tree.
    """

    def __init__(self, data: Union[Iterator, Generator[dict, None, None], Data] = None, ds=None):
        if data is None:
            data = []

        self._data_source = ds
        self._nodes = []
        self.roots: List[TreeNode] = []
        self.events_ids = []
        self.parent_events_ids = set()

        self._build_tree(data)

    def _build_tree(self, data: Union[Iterator, Generator[dict, None, None]]) -> None:
        """Build or append new events to family tree.

        :param data: Events.
        """
        for event in data:
            event = event.copy()

            event_id = event["eventId"]
            try:
                event.pop("body")
            except KeyError:
                pass

            self._nodes.append(TreeNode(name=event["eventName"], data=event))
            self.events_ids.append(event_id)
            self.parent_events_ids.add(event["parentEventId"])

        unknown_parents_ids: list = self._get_unknown_parents_ids()

        # restore them
        restored_events = self._get_unknown_events(unknown_parents_ids)

        for event in restored_events:
            event = event.copy()

            try:
                event.pop("body")
            except KeyError:
                pass

            self._nodes.append(TreeNode(name=event["eventName"], data=event))

        # search roots
        node: Node
        for node in self._nodes.copy():
            if node.data["parentEventId"] is None:
                self.roots.append(node)
                self._nodes.remove(node)

        def x(roots):
            if len(self._nodes) != 0:
                new_roots = []
                for node in self._nodes.copy():
                    for root_node in roots:
                        if node.data["parentEventId"] == root_node.data["eventId"]:
                            node.parent = root_node
                            self._nodes.remove(node)
                            new_roots.append(node)
                x(new_roots)
            else:
                return None

        x(self.roots)

    def _get_unknown_events(self, unknown_parents):
        if unknown_parents:
            new_events: list = self._data_source.find_events_by_id_from_data_provider(unknown_parents)

            unknown_parents2 = set()
            new_events = [new_events] if not isinstance(new_events, list) else new_events
            for e in new_events:
                parent_event_id = e["parentEventId"]
                if parent_event_id is not None:
                    unknown_parents2.add(parent_event_id)

            new_events += self._get_unknown_events(unknown_parents2)

            return new_events
        else:
            return []

    def _get_unknown_parents_ids(self) -> list:
        """Searches unknown events.

        :return: Unknown events.
        """
        unknown_parents_ids = []

        for parent_id in self.parent_events_ids:
            if parent_id is not None:
                if parent_id not in self.events_ids:
                    unknown_parents_ids.append(parent_id)

        return unknown_parents_ids
